from abc import ABC, abstractmethod
from typing import TypeVar, Generic

from griff.utils.errors import BaseError

E = TypeVar("E", bound=BaseError)


class PolicyException(Generic[E], Exception):
    def __init__(self, error: E) -> None:
        self.error = error


class Policy(ABC):
    @abstractmethod
    async def check(self, **kwargs) -> None:
        ...



import abc
from contextlib import asynccontextmanager
from typing import List, TypeVar, Generic, Dict, Any

C = TypeVar("C")

DbRow = Dict[str, Any]


class DbProvider(Generic[C], abc.ABC):
    @abc.abstractmethod
    async def start(self) -> None:  # pragma: no cover
        ...

    @abc.abstractmethod
    async def stop(self) -> None:  # pragma: no cover
        ...

    @abc.abstractmethod
    async def get_connection(self) -> C:  # pragma: no cover
        ...

    @abc.abstractmethod
    async def close_connection(self, connection: C) -> None:  # pragma: no cover
        ...

    @abc.abstractmethod
    async def execute(
        self, connection: C, sql: str | List[str]
    ) -> None:  # pragma: no cover
        ...

    @abc.abstractmethod
    async def fetch_one(self, connection: C, sql: str) -> DbRow:  # pragma: no cover
        ...

    @abc.abstractmethod
    async def fetch_all(
        self, connection: C, sql: str
    ) -> List[DbRow]:  # pragma: no cover
        ...

    @abc.abstractmethod
    async def start_transaction(self, connection: C) -> None:  # pragma: no cover
        ...

    @abc.abstractmethod
    async def commit_transaction(self, connection: C) -> None:  # pragma: no cover
        ...

    @abc.abstractmethod
    async def rollback_transaction(self, connection: C) -> None:  # pragma: no cover
        ...

    @abc.abstractmethod  # type: ignore
    @asynccontextmanager
    async def transaction(self, connection: C):  # pragma: no cover
        ...



from gettext import gettext as _

ExceptionMsg = str | None
ExceptionDetails = dict | list | None


class GriffException(Exception):
    """
    Base class for exceptions

    Attributes:
        code (int): code like HTTP status code
        default_message (str): default error message
        details (dict): additional details
    """

    status_code = 500
    default_message: str = _("error message")

    def __init__(
        self, message: ExceptionMsg = None, details: ExceptionDetails = None
    ) -> None:
        self.message = message or self.default_message
        self.code = self.status_code
        self.details = details or None

    def __str__(self) -> str:
        return f"{self.message}"

    def get_message(self):  # pragma: no cover
        return self.message

    def get_code(self):  # pragma: no cover
        return self.code

    def get_details(self):  # pragma: no cover
        return self.details

    def to_dict(self) -> dict:  # pragma: no cover
        a_dict = {"code": self.code, "message": self.get_message()}
        if self.details is not None:
            a_dict["details"] = self.details
        return a_dict

    @classmethod
    def classname(cls) -> str:  # pragma: no cover
        return str(cls)

    @classmethod
    def short_classname(cls) -> str:  # pragma: no cover
        return cls.__name__


class NotFoundException(GriffException):
    """Raised when something is not found"""

    status_code = 404


class AlreadyExistsException(GriffException):
    """Raised when something already exists"""

    status_code = 409
    default_message = _("Entity or Aggregate already exists")


class ValidationException(GriffException):
    """Raised when data validation failed"""

    status_code = 422
    default_message = _("Unprocessable Entity")


class UnauthorizedException(GriffException):
    """Raised when Authorization failed"""

    status_code = 401
    default_message = _("Unauthorized")


class AuthenticationTimeoutException(GriffException):
    """Raise when Authentification Timeout"""

    status_code = 403
    default_message = _("Authentication Timeout")


class AccessForbiddenException(GriffException):
    """Raise when access is forbidden"""

    status_code = 403
    default_message = _("Access Forbidden")


class BadRequestException(GriffException):
    """
    Raise on an apparent client (e.g., malformed request syntax, size too large, ...)
    """

    status_code = 400
    default_message = _("Bad Request")


class ServiceUnavailableException(GriffException):
    """
    Raise when cannot handle the request (because it is overloaded or down for
    maintenance). Generally, this is a temporary state
    """

    status_code = 503
    default_message = _("Service temporarily unavailable, try again later.")


class EntityNotFoundException(NotFoundException):
    default_message = _("Entity not found")



from abc import ABC

from griff.test_utils.mixins.db_test_mixin import DbTestMixin
from griff.test_utils.pytest_runtime_factory import PytestRuntimeFactory


class DbInTransactionTestMixin(DbTestMixin, ABC):
    @classmethod
    def runtime_factory(cls) -> PytestRuntimeFactory:
        return super().runtime_factory().test_with_db().with_test_in_db_transaction()



from abc import ABC, abstractmethod
from typing import Type, TypeVar, Generic

from griff.appli.query.query_handler import QueryHandler, QueryResponse
from griff.test_utils.mixins.runtime_test_mixin import RuntimeTestMixin
from griff.test_utils.pytest_runtime_factory import PytestRuntimeFactory
from griff.test_utils.testcases.testcase import TestCase

QH = TypeVar("QH", bound=QueryHandler)


class QueryHandlerTestCase(Generic[QH], RuntimeTestMixin, TestCase, ABC):
    handler: QH

    @classmethod
    def runtime_factory(cls) -> PytestRuntimeFactory:
        return super().runtime_factory().command_test_handler()

    def setup_method(self):
        super().setup_method()
        self.handler = self.get_injected(self.handler_class)

    @staticmethod
    def prepare_success_resultset(response: QueryResponse):
        return response.model_dump()

    @property
    @abstractmethod
    def handler_class(self) -> Type[QH]:
        pass



from abc import ABC, abstractmethod

from griff.appli.app_event.app_event import AppEvent
from griff.appli.message.message_middleware import MessageMiddleware, MessageContext


class AppEventMiddleware(MessageMiddleware[AppEvent, None], ABC):
    @abstractmethod
    async def dispatch(
        self, message: AppEvent, context: MessageContext | None = None
    ) -> None:
        pass



import ast
import glob
import importlib.util
from pathlib import Path

from injector import Binder

from griff.runtime.components.abstract_runtime_component import (
    InjectBindable,
    RuntimeComponent,
)
from griff.settings.griff_settings import GriffSettings


def get_classes_from_file(filepath):
    with open(filepath, "r", encoding="utf-8") as file:
        root = ast.parse(file.read())

    classes = [node.name for node in ast.walk(root) if isinstance(node, ast.ClassDef)]
    return classes


def scan_persistence_classes(directory, type_name: str):
    python_classes = []

    for filepath in glob.glob(f"{directory}/*_{type_name}.py", recursive=True):
        classes = get_classes_from_file(filepath)
        if classes:
            class1, class2 = classes
            python_classes.append(
                {
                    "module": str(Path(filepath).name)[:-3],
                    "real": class1 if "Fake" not in class1 else class2,
                    "fake": class1 if "Fake" in class1 else class2,
                }
            )
    return python_classes


class FakePersistenceTestContextRuntimeComponent(InjectBindable, RuntimeComponent):
    def __init__(self, context: str, settings: GriffSettings):
        self.base_module = str(
            settings.get_repositories_path(context=context, absolute=False)
        ).replace("/", ".")
        self.repository_path = settings.get_repositories_path(context=context)

    def configure(self, binder: Binder) -> None:
        self._bind_fake_classes(binder, type_name="persistence")
        self._bind_fake_classes(binder, type_name="repository")

    def _bind_fake_classes(self, binder: Binder, type_name: str):
        for klasses in scan_persistence_classes(self.repository_path, type_name):
            real_klass, fake_klass = self._import_module(klasses)
            binder.bind(real_klass, to=fake_klass)

    def _import_module(self, klasses: dict):
        module_name = f"{self.base_module}.{klasses['module']}"
        imported_module = importlib.import_module(module_name)
        for klass in [klasses["real"], klasses["fake"]]:
            if module_name in globals() and hasattr(globals()[module_name], klass):
                continue
            if hasattr(imported_module, klass) is False:
                raise RuntimeError(f"package {klass} not found in {module_name}")
            globals().update({klass: getattr(imported_module, klass)})
        return getattr(imported_module, klasses["real"]), getattr(
            imported_module, klasses["fake"]
        )



from abc import ABC, abstractmethod

from griff.appli.command.command import Command
from griff.appli.command.command_handler import CommandResponse
from griff.appli.message.message_middleware import MessageMiddleware, MessageContext


class CommandMiddleware(MessageMiddleware[Command, CommandResponse], ABC):
    @abstractmethod
    async def dispatch(
        self, message: Command, context: MessageContext | None = None
    ) -> CommandResponse:
        pass



from abc import ABC, abstractmethod
from inspect import getmembers
from types import FunctionType
from typing import Any, Iterator

import typer
from typer import Typer

from griff.infra.cli.typer_print import TyperPrint


class CliController(ABC):
    def __init__(self) -> None:
        self._app = self._build_app()

    def get_app(self) -> Typer:
        return self._app

    @abstractmethod
    def get_command_name(self) -> str:
        raise NotImplementedError

    def _build_app(self) -> Typer:
        app = Typer()
        self._register_endpoints(app)
        return app

    def _register_endpoints(self, app: Typer) -> None:
        cls = type(self)

        for endpoint in cls._get_endpoints():
            name = getattr(endpoint, "_endpoint_name")
            method = endpoint.__get__(self, cls)
            app.command(name)(method)

    @classmethod
    def _get_endpoints(cls) -> Iterator[FunctionType]:
        def is_endpoint(member: Any) -> bool:
            return isinstance(member, FunctionType) and hasattr(
                member, "_endpoint_name"
            )

        return (value for _, value in getmembers(cls, is_endpoint))

    @staticmethod
    def _check_process_run(p: Any, action: str) -> None:
        if p.returncode != 0:
            TyperPrint.error(f"{action} has failed")
            raise typer.Exit()



from abc import ABC
from dataclasses import dataclass
from typing import Type, Any

from griff.infra.registry.endpoint_decorator_processor import (
    EndpointDecoratorProcessor,
)
from griff.infra.registry.meta_registry import AbstractMetaRegistry
from griff.utils.inspect_utils import get_decorators


@dataclass
class EndpointDefinition:
    endpoint: Any
    http_method: str
    http_route: str
    http_success_code: int


class MetaEndpointControllerRegistry(AbstractMetaRegistry):
    ENDPOINT_REGISTRY: dict = {}

    def __new__(mcs, name, bases, attrs):
        new_cls = type.__new__(mcs, name, bases, attrs)
        if ABC in bases:
            return new_cls

        for key, val in attrs.items():
            if key not in [
                "__module__",
                "__qualname__",
                "base_route",
                "router",
                "__classcell__",
                "__init__",
            ]:
                deco_list = get_decorators(val)

                if deco_list is None or len(deco_list) == 0:
                    continue

                decorator_processor = EndpointDecoratorProcessor()
                decorator_processor.load_decorators(deco_list)

                if decorator_processor.route is None:
                    # decorator is not an endpoint decorator
                    continue

                endpoint_def = EndpointDefinition(
                    endpoint=str(val).split(".")[1].split(" ")[0],
                    http_method=decorator_processor.method,
                    http_route=decorator_processor.route,
                    http_success_code=decorator_processor.code,
                )
                mcs.add_endpoint_to_controller(
                    endpoint=endpoint_def, controller=new_cls
                )

        return new_cls

    @classmethod
    def add_endpoint_to_controller(mcs, endpoint: EndpointDefinition, controller: Type):
        if controller not in mcs.ENDPOINT_REGISTRY.keys():
            mcs.ENDPOINT_REGISTRY[controller] = list()

        if endpoint.http_route in [
            e.http_route for e in mcs.ENDPOINT_REGISTRY[controller]
        ]:
            raise ValueError(
                f"route {endpoint.http_route} already registered "
                f"for {controller.__name__}"
            )

        mcs.ENDPOINT_REGISTRY[controller].append(endpoint)

    @classmethod
    def get_endpoint_registry(mcs):
        return mcs.ENDPOINT_REGISTRY

    @classmethod
    def get_full_route_for_endpoint(cls, controller, endpoint: EndpointDefinition):
        base_route = controller.base_route.strip("/") if controller.base_route else ""
        endpoint_route = endpoint.http_route.strip("/")
        if base_route == "":
            return f"/{endpoint_route.strip('/')}"
        fullroute = f"/{base_route}/{endpoint_route}"
        return fullroute



import subprocess
from pathlib import Path

import typer
from injector import inject

from griff.infra.cli.cli_controller import (
    CliController,
)
from griff.infra.cli.register_cli_command import register_cli_command
from griff.infra.cli.typer_print import TyperPrint
from griff.services.path.path_service import PathService
from griff.settings.griff_settings import GriffSettings


class I18nCliController(CliController):
    @inject
    def __init__(self, settings: GriffSettings):
        super().__init__()
        self._settings = settings
        if settings.locale is None:
            raise ValueError("Locale settings are not defined")
        self._locale_settings = settings.locale

    def get_command_name(self):
        return "i18n"

    @register_cli_command(name="create")
    def init(self):
        TyperPrint.info(f"Initialize locale in {self._locale_settings.relative_path}")
        path_service = PathService()
        path_service.create_missing(self._settings.full_locale_path)
        path_service.create_missing(self._settings.pydantic_locale_path)
        path_service.create_missing(self._locale_settings.fastapi_locale_dir_name)
        p = self._extract_messages()
        self._check_process_run(p, "initialization")
        TyperPrint.success("Locale has been initialized")

    @register_cli_command(name="add_lang")
    def add_lang(self, lang: str):
        TyperPrint.info(f"Adding lang '{lang}'")
        self._check_pydantic_lang_exists(lang)
        pot_filename = self._check_pot_filename_exists()
        p = subprocess.run(
            f"pybabel init -l {lang} -i {pot_filename} -d "
            f"{self._locale_settings.relative_path}",
            shell=True,
        )
        self._check_process_run(p, "adding lang")
        TyperPrint.success(f"Lang '{lang}' has been added")

    @register_cli_command(name="compilemessages")
    def compile(self):
        TyperPrint.info("Translations compiling")
        p = subprocess.run(
            f"pybabel compile -d {self._settings.fastapi_local_path}", shell=True
        )
        self._check_process_run(p, "compilation")
        p = subprocess.run(
            f"pybabel compile -d {self._settings.pydantic_locale_path}",
            shell=True,
        )
        self._check_process_run(p, "compilation pydantic")
        TyperPrint.success("Translation messages have been compiled")

    @register_cli_command(name="update")
    def update(self):
        TyperPrint.info("Translations updating")
        self._extract_messages()
        pot_filename = self._check_pot_filename_exists()
        p = subprocess.run(
            f"pybabel update -i {pot_filename} -d "
            f"{self._locale_settings.relative_path}",
            shell=True,
        )
        self._check_process_run(p, "updating")
        TyperPrint.success("Translation messages have been updated")

    def _extract_messages(self):
        babel_config = f"{self._settings.project_dir}/babel.cfg"
        p = subprocess.run(
            f"pybabel extract . -F {babel_config} -o {self._get_messages_filename()}",
            shell=True,
        )
        return p

    def _get_messages_filename(self) -> Path:
        return Path(self._settings.project_dir).joinpath(
            self._locale_settings.relative_path, "messages.pot"
        )

    def _check_pydantic_lang_exists(self, lang):
        pydantic_lang_path = Path(self._settings.project_dir).joinpath(
            self._settings.pydantic_locale_path, lang.lower()
        )
        if pydantic_lang_path.exists() is False:
            TyperPrint.error(
                f"Pydantic lang '{lang}' is not found, see add new lang section in"
                f" griff/infrastructure/pydantic/locale/README.md"
            )
            raise typer.Exit()

    def _check_pot_filename_exists(self) -> Path:
        pot_filename = self._get_messages_filename()
        if pot_filename.exists():
            return pot_filename
        TyperPrint.error("locale is not initialized: run 'python commands locale init'")
        raise typer.Exit()



from pathlib import Path

import asyncpg

from griff.services.db.providers.asyncpg_provider import AsyncPgProvider
from griff.settings.griff_settings import GriffSettings
from griff.utils.migration_utils import MigrationUtils


# noinspection SqlInjection
class DbTplUtils:  # pragma: no cover
    def __init__(
        self,
        context: str,
        settings: GriffSettings,
    ):
        self.context = context
        self.settings = settings
        name = self.settings.project_name.replace(" ", "_").lower()
        self.tpl_db_name = f"test_{name}_{self.context}"
        self._db_tpl_settings = self.settings.db.model_copy(
            update={"name": self.tpl_db_name}, deep=True
        )
        self.default_db_name = self.settings.db.name
        self._template1_provider = self._get_template1_provider()
        self._db_tpl_provider = self._get_db_tpl_provider()

    @property
    def db_tpl_path(self) -> Path:
        return self.settings.get_test_utils_path(self.context).joinpath("db_tpl")

    @property
    def migrations_path(self) -> Path:
        return self.settings.get_migrations_path(self.context)

    @property
    def dump_filename(self) -> Path:
        return self.db_tpl_path.joinpath("db_tpl.dump")

    @property
    def dump_sql_filename(self) -> Path:
        return self.db_tpl_path.joinpath("db_tpl.sql")

    async def create(self) -> None:
        async with self._template1_provider.connection() as conn:
            await self._template1_provider.execute(conn, self._create_sql)

    async def delete(self) -> None:
        async with self._template1_provider.connection() as conn:
            await self._delete(conn)

    async def is_db_exists(self) -> bool:
        try:
            conn = await self._db_tpl_provider.get_connection()
        except asyncpg.InvalidCatalogNameError:
            return False
        await conn.close()
        return True

    async def recreate(self) -> None:
        async with self._template1_provider.connection() as conn:
            await self._delete(conn)
            await self._template1_provider.execute(conn, self._create_sql)

    async def load(self) -> None:
        async with self._template1_provider.connection() as conn:
            # noinspection PyBroadException
            try:
                await self._template1_provider.execute(conn, self._clone_db_sql)
            except Exception:
                await self._template1_provider.execute(conn, self._delete_cloned_sql)
                await self._template1_provider.execute(conn, self._clone_db_sql)

    async def clean(self) -> None:
        async with self._template1_provider.connection() as conn:
            await self._template1_provider.execute(conn, self._delete_cloned_sql)

    def is_context_exists(self) -> bool:
        context_path = self.settings.get_context_path(self.context)
        return context_path.exists()

    async def restore_from_sql(self) -> None:
        if self.dump_sql_filename.exists() is False:
            raise RuntimeError(f"no dump sql found for context {self.context}")
        with open(str(self.dump_sql_filename)) as f:
            sql = f.read()

        await self.recreate()

        async with self._db_tpl_provider.connection() as conn:
            await self._db_tpl_provider.execute(conn, sql)

    def do_migration(self) -> None:
        if self.migrations_path.exists() is False:
            raise RuntimeError(f"Migration path for context '{self.context}' not found")
        MigrationUtils.migrate(self._db_tpl_settings.dsn, str(self.migrations_path))

    async def _delete(self, conn) -> None:
        await self._template1_provider.disconnect_all(conn)
        # drop table must be executed alone
        await self._template1_provider.execute(conn, self._delete_sql)

    def _get_template1_provider(self) -> AsyncPgProvider:
        settings = self.settings.db.model_copy(update={"name": "template1"}, deep=True)
        return AsyncPgProvider(settings)

    def _get_db_tpl_provider(self) -> AsyncPgProvider:
        return AsyncPgProvider(self._db_tpl_settings)

    @property
    def _initial_data_path(self) -> Path:
        return self.db_tpl_path.joinpath("initial_db_tpl_data.py")

    @property
    def _create_sql(self) -> str:
        return f"CREATE DATABASE {self.tpl_db_name}"

    @property
    def _delete_sql(self) -> str:
        return f"DROP DATABASE IF EXISTS {self.tpl_db_name}"

    @property
    def _clone_db_sql(self) -> str:
        return f"CREATE DATABASE {self.default_db_name} TEMPLATE {self.tpl_db_name}"

    @property
    def _delete_cloned_sql(self) -> str:
        return f"DROP DATABASE {self.default_db_name}"



from typing import Self, TypeVar, Generic, Type

from griff.appli.command.command_middleware import CommandMiddleware
from griff.appli.command.middlewares.command_event_dispatch_middleware import (
    CommandEventDispatchMiddleware,
)
from griff.appli.event.event_middleware import EventMiddleware
from griff.appli.query.query_middleware import QueryMiddleware
from griff.context_entry_point import ContextEntryPoint
from griff.runtime.components.db_runtime_component import DbTestRuntimeComponent
from griff.runtime.components.db_test_context_in_transaction_runtime_component import (
    DbTestContextInTransactionRuntimeComponent,
)
from griff.runtime.components.db_test_in_transaction_component import (
    DbTestInTransactionRuntimeComponent,
)
from griff.runtime.components.default_services_runtime_component import (
    DefaultServicesTestRuntimeComponent,
)
from griff.runtime.components.fake_persistence_test_context_runtime_component import (
    FakePersistenceTestContextRuntimeComponent,
)
from griff.runtime.factories.abstract_runtime_factory import (
    AbstractRuntimeFactory,
    Injectables,
)
from griff.settings.pytest_settings import PytestSettings

S = TypeVar("S", bound=PytestSettings)


class PytestRuntimeFactory(Generic[S], AbstractRuntimeFactory[S]):  # pragma: no cover
    def base_test(self) -> Self:
        self.with_log_level().with_default_services().with_service_locator()
        return self

    def test_with_db(self) -> Self:
        return self.base_test().with_db_test()

    def domain_test(self) -> Self:
        return self.base_test()

    def repository_test(
        self, context_entry_point_class: Type[ContextEntryPoint]
    ) -> Self:
        return self.test_with_db().with_context_in_db_transaction(
            context_entry_point_class.context_name()
        )

    def persistence_test(
        self, context_entry_point_class: Type[ContextEntryPoint]
    ) -> Self:
        return self.test_with_db().with_context_in_db_transaction(
            context_entry_point_class.context_name()
        )

    def api_test(self, context_entry_point_class: Type[ContextEntryPoint]) -> Self:
        return (
            self.test_with_db()
            .with_context_in_db_transaction(context_entry_point_class.context_name())
            .with_command_bus()
            .with_event_bus()
            .with_query_bus()
            .with_fastapi()
            .with_context_api(context_entry_point_class)
        )

    def cli_test(self, context_entry_point_class: Type[ContextEntryPoint]) -> Self:
        return (
            self.test_with_db()
            .with_context_in_db_transaction(context_entry_point_class.context_name())
            .with_command_bus(for_cli=True)
            .with_event_bus(for_cli=True)
            .with_query_bus(for_cli=True)
            .with_context_api(context_entry_point_class)
        )

    def command_test_handler(self) -> Self:
        return self.base_test()

    def event_test_handler(self) -> Self:
        return self.base_test()

    def with_default_services(self) -> Self:
        if self._settings is None:
            # to avoid mypy error settings can be None
            raise self._settings_missing_exception()
        self._add_component(DefaultServicesTestRuntimeComponent(self._settings))
        return self

    def with_injectables(self, injectables: Injectables | None = None) -> Self:
        if injectables is None:
            return self
        return super().with_injectables(injectables)

    def with_db_test(self) -> Self:
        if self._settings is None:
            # to avoid mypy error settings can be None
            raise self._settings_missing_exception()
        self._add_component(DbTestRuntimeComponent(self._settings))
        return self

    def with_test_in_db_transaction(self) -> Self:
        if self._settings is None:
            # to avoid mypy error settings can be None
            raise self._settings_missing_exception()
        self._add_component(DbTestInTransactionRuntimeComponent())
        return self

    def with_context_in_db_transaction(self, context_name: str) -> Self:
        if self._settings is None:
            # to avoid mypy error settings can be None
            raise self._settings_missing_exception()
        self._add_component(
            DbTestContextInTransactionRuntimeComponent(context_name, self._settings)
        )
        return self

    def with_context_fake_persistence(self, context_name: str) -> Self:
        if self._settings is None:
            # to avoid mypy error settings can be None
            raise self._settings_missing_exception()
        self._add_component(
            FakePersistenceTestContextRuntimeComponent(context_name, self._settings)
        )
        return self

    def _command_bus_middlewares(self) -> list[Type[CommandMiddleware]]:
        return [CommandEventDispatchMiddleware]

    # noinspection PyMethodMayBeStatic
    def _event_bus_middlewares(self) -> list[Type[EventMiddleware]]:
        return []

    # noinspection PyMethodMayBeStatic
    def _query_bus_middlewares(self) -> list[Type[QueryMiddleware]]:
        return []



import copy
from typing import Dict, List, Callable, Any

from griff.infra.persistence.persistence import (
    Persistence,
    QueryRowResult,
    QueryResult,
    QueryRowResults,
)
from griff.infra.persistence.serialized_persistence import SerializedPersistence
from griff.services.json.json_service import JsonService

QueryName = str
QueryCallable = Callable[..., QueryResult]


def _has_filters_match(data: Dict[str, Dict], filters: Dict[str, Any]) -> bool:
    for attr, value in filters.items():
        if data.get(attr) != value:
            return False
    return True


class DictPersistence(Persistence):
    def __init__(self, initial_data: List[Dict] | None = None):
        self._internal_storage: Dict[str, Dict] = {}
        self.reset(initial_data)

    async def _insert(self, data: dict) -> None:
        if data["entity_id"] not in self._internal_storage:
            self._internal_storage[data["entity_id"]] = data
            return None
        raise ValueError(f"id '{data['entity_id']}' already exists")

    async def _update(self, data: dict) -> None:
        if data["entity_id"] in self._internal_storage:
            self._internal_storage[data["entity_id"]] = data
            return None
        raise ValueError(f"id '{data['entity_id']}' does not exists")

    async def _delete(self, persistence_id: str) -> None:
        if persistence_id in self._internal_storage:
            self._internal_storage.pop(persistence_id)
            return None
        raise ValueError(f"id '{persistence_id}' does not exists")

    async def _get_by_id(self, persistence_id: str) -> QueryRowResult:
        if persistence_id in self._internal_storage:
            return copy.deepcopy(self._internal_storage[persistence_id])
        raise ValueError(f"id '{persistence_id}' not found")

    async def _list_all(self) -> QueryRowResults:
        return copy.deepcopy(list(self._internal_storage.values()))

    async def _run_query(self, query_name: str, **query_params) -> QueryResult:
        if self._has_custom_queries(query_name):
            return self._run_custom_queries(query_name, **query_params)
        if "get_by_" in query_name:
            return self._get_by_attrs(**query_params)
        if "list_by_" in query_name:
            return self._list_by_attrs(**query_params)
        raise RuntimeError(f"Query {query_name} not found")

    def reset(self, initial_data: List[Dict] | None = None):
        if initial_data is None:
            self._internal_storage = {}
            return None
        self._internal_storage = {
            e["entity_id"]: self._prepare_to_save(copy.deepcopy(e))
            for e in initial_data
        }

    @property
    def _queries(self) -> Dict[QueryName, QueryCallable]:
        return {}

    def _has_custom_queries(self, query_name: str) -> bool:
        return query_name in self._queries

    def _run_custom_queries(self, query_name: str, **query_params):
        return self._queries[query_name](**query_params)

    def _searchable_internal_storage(self):
        if self._internal_storage:
            return self._prepare_row_results(
                [copy.deepcopy(r) for r in self._internal_storage.values()]
            )
        return self._internal_storage  # pragma: no cover

    def _get_by_attrs(
        self, filtering_callable: Callable = _has_filters_match, **query_params
    ):
        for e in self._searchable_internal_storage():
            if filtering_callable(e, query_params):
                return self._internal_storage[e["entity_id"]]
        return None

    def _list_by_attrs(
        self, filtering_callable: Callable = _has_filters_match, **query_params
    ):
        result = []
        for e in self._searchable_internal_storage():
            if filtering_callable(e, query_params):
                result.append(self._internal_storage[e["entity_id"]])
        return result


class SerializedDictPersistence(SerializedPersistence, DictPersistence):
    def __init__(self, initial_data: List[Dict] | None = None):
        SerializedPersistence.__init__(self, JsonService())
        DictPersistence.__init__(self, initial_data)



from typing import Type

from injector import Injector, Binder, singleton

from griff.appli.app_event.app_event_bus import AppEventBus
from griff.appli.event.event_middleware import EventMiddleware
from griff.appli.message.message_bus import ListMiddlewares
from griff.infra.registry.meta_registry import MetaAppEventHandlerRegistry
from griff.runtime.components.abstract_runtime_component import (
    RuntimeComponent,
    Runnable,
    InjectBindable,
    import_handlers,
)
from griff.settings.griff_settings import GriffSettings


class AppEventBusRuntimeComponent(Runnable, InjectBindable, RuntimeComponent):
    def __init__(
        self, settings: GriffSettings, middlewares: list[Type[EventMiddleware]]
    ):
        self._settings = settings
        self._middlewares = middlewares

    def configure(self, binder: Binder) -> None:
        binder.bind(AppEventBus, to=AppEventBus, scope=singleton)

    def initialize(self, injector: Injector):
        app_event_bus = injector.get(AppEventBus)
        middlewares: ListMiddlewares = [injector.get(m) for m in self._middlewares]
        for context in self._settings.bounded_contexts:
            import_handlers(
                self._settings.project_dir,
                self._settings.get_app_event_handlers_path(context),
            )

        handlers = [injector.get(h) for h in MetaAppEventHandlerRegistry.list_types()]
        app_event_bus.initialize(handlers, middlewares)

    def clean(self, injector: Injector):
        pass

    def start(self, injector: Injector):
        pass

    def stop(self, injector: Injector):
        pass



from injector import inject

from griff.appli.message.message_bus import MessageBus
from griff.appli.query.query import Query
from griff.appli.query.query_dispatcher import QueryDispatcher
from griff.appli.query.query_handler import QueryResponse, QueryHandler


class QueryBus(MessageBus[Query, QueryResponse, QueryHandler]):
    @inject
    def __init__(self, dispatcher: QueryDispatcher) -> None:
        super().__init__(dispatcher)



from abc import ABC, abstractmethod
from typing import Type, TypeVar, Generic

from griff.appli.command.command_handler import CommandHandler, CommandResponse
from griff.infra.persistence.dict_persistence import DictPersistence
from griff.infra.repository.repository import Repository
from griff.test_utils.mixins.runtime_test_mixin import RuntimeTestMixin
from griff.test_utils.pytest_runtime_factory import PytestRuntimeFactory
from griff.test_utils.testcases.testcase import TestCase

CH = TypeVar("CH", bound=CommandHandler)


class CommandHandlerTestCase(Generic[CH], RuntimeTestMixin, TestCase, ABC):
    handler: CH

    @classmethod
    def runtime_factory(cls) -> PytestRuntimeFactory:
        return super().runtime_factory().command_test_handler()

    def setup_method(self):
        super().setup_method()
        self.handler = self.get_injected(self.handler_class)

    @staticmethod
    def prepare_success_resultset(response: CommandResponse, repository: Repository):
        # noinspection PyTypeChecker
        persistence: DictPersistence = repository._persistence  # type: ignore
        return {
            "response": response.model_dump(),
            "persistence": persistence._searchable_internal_storage(),
        }

    @property
    @abstractmethod
    def handler_class(self) -> Type[CH]:
        pass



from pydantic import BaseModel, ConfigDict


class JwtTokens(BaseModel):
    model_config = ConfigDict(frozen=True)

    access_token: str
    refresh_token: str


class DecodedJwtToken(BaseModel):
    payload: dict
    has_expired: bool



from abc import ABC, abstractmethod
from typing import TypeVar, Generic

from griff.infra.api.api_router import ApiRouter
from griff.infra.cli.cli_router import CliRouter
from griff.infra.registry.meta_registry import MetaContextEntryPointRegistry

AR = TypeVar("AR", bound=ApiRouter)
CR = TypeVar("CR", bound=CliRouter)


class ContextEntryPoint(Generic[AR, CR], ABC, metaclass=MetaContextEntryPointRegistry):
    @abstractmethod
    def __init__(self, api_router: AR | None = None, cli_router: CR | None = None):
        self._api_router = api_router
        self._cli_router = cli_router

    def get_api_router(self) -> AR | None:  # pragma: no cover
        return self._api_router

    def get_cli_router(self) -> CR | None:  # pragma: no cover
        return self._cli_router

    @staticmethod
    @abstractmethod
    def context_name() -> str:  # pragma: no cover
        ...



import copy
from abc import ABC, abstractmethod
from typing import Type, List

from griff.context_entry_point import ContextEntryPoint
from griff.infra.persistence.persistence import Persistence
from griff.test_utils.mixins.runtime_test_mixin import RuntimeTestMixin
from griff.test_utils.pytest_runtime_factory import PytestRuntimeFactory
from griff.test_utils.testcases.testcase import TestCase


class PersistenceTestCase(RuntimeTestMixin, TestCase, ABC):
    @classmethod
    @abstractmethod
    def entry_point_class(cls) -> Type[ContextEntryPoint]:
        pass

    @classmethod
    def runtime_factory(cls) -> PytestRuntimeFactory:
        return super().runtime_factory().persistence_test(cls.entry_point_class())

    def setup_method(self):
        super().setup_method()

    async def run_crud(
        self,
        persistence: Persistence,
        rows: List[dict],
        row_to_update: dict,
        update_data: dict,
        row_to_delete: dict,
    ):
        # insert, read
        for row in rows:
            await persistence.insert(row)
        inserted_data = await persistence.list_all()
        # update, read
        before_update = row_to_update
        updated_row = {**copy.deepcopy(before_update), **update_data}
        await persistence.update(updated_row)
        after_update = await persistence.get_by_id(before_update["entity_id"])
        # delete, read
        await persistence.delete(row_to_delete["entity_id"])
        assert await persistence.get_by_id(row_to_delete["entity_id"]) is None
        return {
            "inserted_data": inserted_data,  # type: ignore
            "after_update": after_update,
        }



import datetime
from typing import Optional

import arrow
from injector import singleton

from griff.services.abstract_service import AbstractService
from griff.services.date.date_models import DateTime


@singleton
class DateService(AbstractService):
    def __init__(self, locale="fr-fr"):
        self._locale = locale

    def now(self) -> DateTime:
        return DateTime(arrow.utcnow())

    def to_mysql_date(self, d: DateTime | None = None) -> str:
        return self._get(d).to_mysql_date()

    def to_mysql_datetime(self, d: DateTime | None = None) -> str:
        return self._get(d).to_mysql_datetime()

    def to_date(self, d: DateTime | None = None) -> datetime.date:
        return self._get(d).to_date()

    def to_datetime(self, d: DateTime | None = None) -> datetime.datetime:
        return self._get(d).to_datetime()

    def _get(self, d: Optional[DateTime] = None):
        if d and isinstance(d, DateTime) is False:
            raise ValueError("Invalid DateTime or Date instance")
        return self.now() if d is None else d



from abc import ABC
from typing import TypeVar, Generic

from griff.appli.app_event.app_event import AppEvent
from griff.appli.message.message_handler import (
    MessageHandler,
)
from griff.infra.registry.meta_registry import (
    MetaAppEventHandlerRegistry,
)

EM = TypeVar("EM", bound=AppEvent)


class AppEventHandler(
    Generic[EM], MessageHandler[EM, None], ABC, metaclass=MetaAppEventHandlerRegistry
):
    ...



from typing import TypeVar

from injector import singleton, Binder

from griff.runtime.components.abstract_runtime_component import (
    InjectBindable,
    RuntimeComponent,
)
from griff.services.date.date_service import DateService
from griff.services.date.fake_date_service import FakeDateService
from griff.services.db.db_settings import DbSettings
from griff.services.hasher.providers.argon2_provider import Argon2Provider
from griff.services.hasher.providers.fake_hasher_provider import FakeHasherProvider
from griff.services.hasher.providers.hasher_provider import HasherProvider
from griff.services.jwt.jwt_settings import JwtSettings
from griff.services.query_runner.query_runner_settings import QueryRunnerSettings
from griff.services.uniqid.generator.fake_uniqid_generator import FakeUniqIdGenerator
from griff.services.uniqid.generator.ulid_uniqid_generator import UlidUniqIdGenerator
from griff.services.uniqid.generator.uniqid_generator import (
    UniqIdGenerator,
)
from griff.settings.griff_settings import GriffSettings

S = TypeVar("S", bound=GriffSettings)


class DefaultServicesRuntimeComponent(InjectBindable, RuntimeComponent):
    def __init__(self, settings: S):
        self._settings = settings

    def configure(self, binder: Binder) -> None:
        binder.bind(UniqIdGenerator, to=UlidUniqIdGenerator, scope=singleton)  # type: ignore # noqa: E501
        binder.bind(HasherProvider, to=Argon2Provider, scope=singleton)  # type: ignore
        binder.bind(JwtSettings, to=self._settings.jwt, scope=singleton)  # type: ignore
        binder.bind(QueryRunnerSettings, to=self._settings.query_runner)
        binder.bind(DbSettings, to=self._settings.db)


class DefaultServicesTestRuntimeComponent(InjectBindable, RuntimeComponent):
    def __init__(self, settings: S):
        self._settings = settings

    def configure(self, binder: Binder) -> None:
        binder.bind(DateService, to=FakeDateService)
        binder.bind(UniqIdGenerator, to=FakeUniqIdGenerator)  # type: ignore # noqa: E501
        binder.bind(HasherProvider, to=FakeHasherProvider)  # type: ignore
        binder.bind(JwtSettings, to=self._settings.jwt)
        binder.bind(QueryRunnerSettings, to=self._settings.query_runner)
        binder.bind(DbSettings, to=self._settings.db)



from typing import Type, List, Any

from injector import Injector, Binder, singleton

from griff.appli.command.command_bus import CommandBus
from griff.appli.command.command_handler import CommandHandler
from griff.appli.command.command_middleware import CommandMiddleware
from griff.appli.message.message_bus import ListMiddlewares
from griff.infra.registry.meta_registry import MetaCommandHandlerRegistry
from griff.runtime.components.abstract_runtime_component import (
    RuntimeComponent,
    Runnable,
    InjectBindable,
)


class CommandBusRuntimeComponent(Runnable, InjectBindable, RuntimeComponent):
    def __init__(self, middlewares: list[Type[CommandMiddleware]]):
        self._middlewares = middlewares

    def configure(self, binder: Binder) -> None:
        binder.bind(CommandBus, to=CommandBus, scope=singleton)

    def initialize(self, injector: Injector) -> None:
        command_bus = injector.get(CommandBus)
        middlewares: ListMiddlewares = [injector.get(m) for m in self._middlewares]
        handlers: List[CommandHandler[Any, Any]] = [
            injector.get(h) for h in MetaCommandHandlerRegistry.list_types()
        ]
        command_bus.initialize(handlers, middlewares)

    def clean(self, injector: Injector) -> None:
        pass

    def start(self, injector: Injector) -> None:
        pass

    def stop(self, injector: Injector) -> None:
        pass



from griff.services.hasher.providers.hasher_provider import (
    HasherProvider,
    UnHashedStr,
    HashedStr,
)


class FakeHasherProvider(HasherProvider):
    def hash(self, to_hash: UnHashedStr) -> HashedStr:
        return to_hash.replace(" ", "µ")

    def verify(self, candidate: UnHashedStr, reference: HashedStr) -> bool:
        return candidate.replace(" ", "µ") == reference



import functools


# noinspection PyPep8Naming
class register_endpoint:
    def __init__(self, route=None, method=None, success_code=None):
        self.route = route
        self.method = method

    def __call__(self, func):
        """Calling the class."""

        @functools.wraps(func)
        async def wrapper(*args, **kwargs):  # pragma: no cover
            value = await func(*args, **kwargs)
            return value

        return wrapper



import json
from _decimal import Decimal
from pathlib import Path
from typing import Any

import orjson
from injector import singleton
from pydantic import BaseModel

from griff.services.abstract_service import AbstractService
from griff.services.path.path_service import PathService


@singleton
class JsonService(AbstractService):
    _path_service = PathService()

    def to_json_dumpable(self, d):
        json_bytes = orjson.dumps(d, default=self._orjson_default)
        return orjson.loads(json_bytes)

    def dump(self, d):
        json_bytes = orjson.dumps(d, default=self._orjson_default)
        return json_bytes.decode("utf-8")

    @staticmethod
    def load_from_str(str_json: str):
        try:
            return json.loads(str_json)
        except json.decoder.JSONDecodeError as e:
            raise ValueError(f"Invalid json: {e}")

    def dump_to_file(self, data, filename, human_readable=False):
        default_args = {"obj": self.to_json_dumpable(data)}
        human_readable_args = {
            "indent": 4,
            "separators": (",", ": "),
            "default": str,
            # disable non-ASCII characters escape with \uXXXX sequences
            "ensure_ascii": False,
        }
        with open(filename, "w") as f:
            args = {**default_args, **{"fp": f}}
            if human_readable:
                args = {**args, **human_readable_args}
            json.dump(**args)

    def load_from_file(self, filename: str | Path) -> Any:
        self._path_service.check_exists(filename)
        with open(filename) as f:
            return json.load(f)

    @staticmethod
    def _orjson_default(data):
        if isinstance(data, Decimal):
            return str(data)

        if isinstance(data, bytes):
            return data.decode("utf-8")

        if isinstance(data, set):
            return list(data)

        if isinstance(data, BaseModel):
            return data.model_dump()

        raise TypeError



import sys
from abc import ABC
from typing import Self, Type, TypeVar, Generic, Any

from loguru import logger

from griff.appli.app_event.middlewares.app_event_logger_middleware import (
    AppEventLoggerMiddleware,
)
from griff.appli.command.command_middleware import CommandMiddleware
from griff.appli.command.middlewares.command_event_dispatch_middleware import (
    CommandEventDispatchMiddleware,
)
from griff.appli.command.middlewares.command_logger_middleware import (
    CommandLoggerMiddleware,
)
from griff.appli.command.middlewares.command_uow_middleware import CommandUowMiddleware
from griff.appli.event.event_middleware import EventMiddleware
from griff.appli.event.middlewares.event_logger_middleware import EventLoggerMiddleware
from griff.appli.query.middlewares.query_logger_middleware import QueryLoggerMiddleware
from griff.appli.query.query_middleware import QueryMiddleware
from griff.context_entry_point import ContextEntryPoint, AR, CR
from griff.runtime.components.abstract_runtime_component import RuntimeComponent
from griff.runtime.components.app_event_bus_runtime_component import (
    AppEventBusRuntimeComponent,
)
from griff.runtime.components.command_bus_runtime_component import (
    CommandBusRuntimeComponent,
)
from griff.runtime.components.common_cli_runtime_component import (
    CommonCliRuntimeComponent,
)
from griff.runtime.components.context_api_runtime_component import (
    ContextApiRuntimeComponent,
)
from griff.runtime.components.context_cli_runtime_component import (
    ContextCliRuntimeComponent,
)
from griff.runtime.components.db_runtime_component import DbRuntimeComponent
from griff.runtime.components.default_services_runtime_component import (
    DefaultServicesRuntimeComponent,
)
from griff.runtime.components.event_bus_runtime_component import (
    EventBusRuntimeComponent,
)
from griff.runtime.components.fastapi_runtime_component import FastApiRuntimeComponent
from griff.runtime.components.inject_runtime_component import (
    InjectRuntimeComponent,
    Injectable,
)
from griff.runtime.components.query_bus_runtime_component import (
    QueryBusRuntimeComponent,
)
from griff.runtime.components.service_locator_runtime_component import (
    ServiceLocatorRuntimeComponent,
)
from griff.runtime.runtime import Runtime
from griff.services.date.date_service import DateService
from griff.services.uniqid.uniqid_service import UniqIdService
from griff.settings.griff_settings import GriffSettings, LogLevel

S = TypeVar("S", bound=GriffSettings)
Injectables = dict[Type[Any], Injectable]


class AbstractRuntimeFactory(Generic[S], ABC):  # pragma: no cover
    def __init__(self, settings: S | None = None) -> None:
        self._settings = settings
        self._components: list[RuntimeComponent] = []
        self.with_injectable(GriffSettings, self._settings)

    def build(self) -> Runtime[S]:
        return self._build()

    def with_injectable(self, klass: Type[Any], to: Injectable) -> Self:
        self._add_component(InjectRuntimeComponent(klass, to))
        return self

    def with_default_services(self) -> Self:
        if self._settings is None:
            # to avoid mypy error settings can be None
            raise self._settings_missing_exception()
        self._add_component(DefaultServicesRuntimeComponent(self._settings))
        return self

    def with_service_locator(self) -> Self:
        self._add_component(
            ServiceLocatorRuntimeComponent([UniqIdService, DateService])
        )
        return self

    def with_injectables(self, injectables: Injectables) -> Self:
        for klass, to in injectables.items():
            self.with_injectable(klass, to)
        return self

    def with_db(self) -> Self:
        if self._settings is None:
            # to avoid mypy error settings can be None
            raise self._settings_missing_exception()
        self._add_component(DbRuntimeComponent(self._settings))
        return self

    def with_log_level(self, log_level: LogLevel | None = None) -> Self:
        logger.remove()
        default_log_level = (
            LogLevel.INFO.value
            if self._settings is None
            else self._settings.log_level.value
        )
        level = default_log_level if log_level is None else log_level
        logger.add(sys.stderr, level=level)
        return self

    def with_command_bus(self, for_cli: bool = False) -> Self:
        middlewares = (
            self._cli_command_bus_middlewares()
            if for_cli
            else self._command_bus_middlewares()
        )
        self._add_component(CommandBusRuntimeComponent(middlewares))
        return self

    def with_event_bus(self, for_cli: bool = False) -> Self:
        if self._settings is None:
            # to avoid mypy error settings can be None
            raise self._settings_missing_exception()
        middlewares = (
            self._cli_event_bus_middlewares()
            if for_cli
            else self._event_bus_middlewares()
        )
        self._add_component(EventBusRuntimeComponent(self._settings, middlewares))
        return self

    def with_app_event_bus(self, for_cli: bool = False) -> Self:
        if self._settings is None:
            # to avoid mypy error settings can be None
            raise self._settings_missing_exception()
        middlewares = (
            self._cli_app_event_bus_middlewares()
            if for_cli
            else self._app_event_bus_middlewares()
        )
        self._add_component(AppEventBusRuntimeComponent(self._settings, middlewares))
        return self

    def with_query_bus(self, for_cli: bool = False) -> Self:
        middlewares = (
            self._cli_query_bus_middlewares()
            if for_cli
            else self._query_bus_middlewares()
        )
        self._add_component(QueryBusRuntimeComponent(middlewares))
        return self

    def with_fastapi(self) -> Self:
        if self._settings is None:
            # to avoid mypy error settings can be None
            raise self._settings_missing_exception()
        self._add_component(FastApiRuntimeComponent(self._settings))
        return self

    def with_context_api(
        self, context_entry_point_class: Type[ContextEntryPoint[AR, CR]]
    ) -> Self:
        if self._settings is None:
            # to avoid mypy error settings can be None
            raise self._settings_missing_exception()
        self._add_component(ContextApiRuntimeComponent(context_entry_point_class))
        return self

    def with_common_cli(self) -> Self:
        if self._settings is None:
            # to avoid mypy error settings can be None
            raise self._settings_missing_exception()
        self._add_component(CommonCliRuntimeComponent())
        return self

    def with_context_cli(
        self, context_entry_point_class: Type[ContextEntryPoint[AR, CR]]
    ) -> Self:
        if self._settings is None:
            # to avoid mypy error settings can be None
            raise self._settings_missing_exception()
        self._add_component(ContextCliRuntimeComponent(context_entry_point_class))
        return self

    def _add_component(self, component: RuntimeComponent) -> Self:
        self._components.append(component)
        return self

    def _build(self) -> Runtime[S]:
        if self._settings is None:
            # to avoid mypy error settings can be None
            raise self._settings_missing_exception()
        return Runtime(self._settings, self._components)

    # noinspection PyMethodMayBeStatic
    def _command_bus_middlewares(self) -> list[Type[CommandMiddleware]]:
        return [
            CommandLoggerMiddleware,
            CommandUowMiddleware,
            CommandEventDispatchMiddleware,
        ]

    # noinspection PyMethodMayBeStatic
    def _event_bus_middlewares(self) -> list[Type[EventMiddleware]]:
        return [EventLoggerMiddleware]

    # noinspection PyMethodMayBeStatic
    def _app_event_bus_middlewares(self) -> list[Type[EventMiddleware]]:
        return [AppEventLoggerMiddleware]

    # noinspection PyMethodMayBeStatic
    def _query_bus_middlewares(self) -> list[Type[QueryMiddleware]]:
        return [QueryLoggerMiddleware]

    def _cli_command_bus_middlewares(self) -> list[Type[CommandMiddleware]]:
        return self._command_bus_middlewares()

    def _cli_event_bus_middlewares(self) -> list[Type[EventMiddleware]]:
        return self._event_bus_middlewares()

    def _cli_app_event_bus_middlewares(self) -> list[Type[EventMiddleware]]:
        return self._app_event_bus_middlewares()

    def _cli_query_bus_middlewares(self) -> list[Type[QueryMiddleware]]:
        return self._query_bus_middlewares()

    @staticmethod
    def _settings_missing_exception() -> RuntimeError:
        return RuntimeError("Settings are missing, call with_settings before with_db")



from injector import inject

from griff.appli.event.event import Event
from griff.appli.event.event_dispatcher import EventDispatcher
from griff.appli.event.event_handler import EventHandler
from griff.appli.message.message_bus import MessageBus


class EventBus(MessageBus[Event, None, EventHandler]):
    @inject
    def __init__(self, dispatcher: EventDispatcher) -> None:
        super().__init__(dispatcher)



from gettext import gettext as _
from pathlib import Path

from pydantic import (
    AfterValidator,
    StringConstraints,
)
from typing_extensions import Annotated


def _check_is_relative_path(path: Path) -> Path:
    if path.exists():
        raise ValueError(_("'%(path)s' is not a relative path") % {"path": path})
    return path


def validate_is_relative_path(v: str) -> str:
    return str(_check_is_relative_path(Path(v)))


def validate_is_relative_dest_path(v: Path) -> Path:
    path = Path(v)
    validate_is_relative_path(str(path.parent))
    return v


def validate_path_exists(v: str) -> str:
    if Path(v).exists():
        return v
    raise ValueError(_("'%(path)s' not found") % {"path": v})


def validate_dest_filename_exists(v: str) -> str:
    path = Path(v)
    validate_path_exists(str(path.parent))
    return v


# String
NoEmptyStr = Annotated[str, StringConstraints(strip_whitespace=True, min_length=1)]

# Path
FilenameStr = Annotated[NoEmptyStr, AfterValidator(validate_path_exists)]
DirectoryStr = FilenameStr
DestFilenameStr = Annotated[NoEmptyStr, AfterValidator(validate_dest_filename_exists)]
RelativeDirectoryStr = Annotated[NoEmptyStr, AfterValidator(validate_is_relative_path)]
RelativeFilenameStr = Annotated[NoEmptyStr, AfterValidator(validate_is_relative_path)]
RelativeDestFilenameStr = Annotated[
    NoEmptyStr, AfterValidator(validate_is_relative_dest_path)
]



from abc import abstractmethod, ABC
from typing import Type, Any

from griff.appli.event.event_handler import FakeEventHandler
from griff.context_entry_point import ContextEntryPoint
from griff.infra.repository.repository import Repository
from griff.test_utils.mixins.runtime_test_mixin import RuntimeTestMixin
from griff.test_utils.pytest_runtime_factory import PytestRuntimeFactory
from griff.test_utils.testcases.testcase import TestCase
from griff.utils.async_utils import AsyncUtils


class CliTestCase(RuntimeTestMixin, TestCase, ABC):
    @classmethod
    @abstractmethod
    def entry_point_class(cls) -> Type[ContextEntryPoint]:
        pass

    @classmethod
    def runtime_factory(cls) -> PytestRuntimeFactory:
        return super().runtime_factory().cli_test(cls.entry_point_class())

    @staticmethod
    def prepare_success_resultset(
        repository: Repository, fake_event_handler: FakeEventHandler | None = None
    ):
        results = AsyncUtils.async_to_sync(repository._persistence.list_all)
        actual: dict[str, Any] = {"persistence": results}
        if fake_event_handler:
            actual["handled_events"] = fake_event_handler.list_events_handled()

        return actual



from injector import singleton

from griff.services.uniqid.generator.uniqid_generator import UniqIdGenerator


@singleton
class FakeUniqIdGenerator(UniqIdGenerator):
    def __init__(self, start_id=1):
        self._start = 0
        self._start = start_id - 1

    def next_id(self) -> str:
        self._start = self._start + 1
        return self._format_id(self._start)

    def reset(self, start_id: int = 1):
        self._start = start_id - 1

    @staticmethod
    def _format_id(str_id: int) -> str:
        return f"FAKEULID{str_id:018d}"



import os

import typer
from injector import inject

from griff.infra.cli.controllers.abstract_db_cli_controller import (
    AbstractDbCliController,
)
from griff.infra.cli.register_cli_command import register_cli_command
from griff.infra.cli.typer_print import TyperPrint
from griff.services.db.providers.asyncpg_provider import AsyncPgProvider
from griff.services.path.path_service import PathService
from griff.settings.griff_settings import GriffSettings
from griff.utils.async_utils import AsyncUtils
from griff.utils.db_tpl_utils import DbTplUtils
from griff.utils.migration_utils import MigrationUtils


class DbCliController(AbstractDbCliController):
    @inject
    def __init__(self, settings: GriffSettings, path_service: PathService):
        super().__init__(settings)
        self._path_service = path_service

    def get_command_name(self):
        return "db"

    def _get_migrations_dir(self, domain_name: str) -> str:
        return str(self._settings.get_migrations_path(domain_name))

    @register_cli_command(name="init-bdd")
    def initialize_bdd(self, noinput: bool = False):
        if noinput is False:
            confirmed = typer.confirm(
                f"It will erase entire database "
                f"'{self._settings.db.name}', are you sure?"
            )
            if not confirmed:
                TyperPrint.warning("Ok, no pb, see you later !!")
                raise typer.Exit()
        TyperPrint.info("Init Bdd")

        AsyncUtils.async_to_sync(self._recreate_db)

        self._do_migrations()

        if self._settings.env in ["dev", "pytest"]:
            AsyncUtils.async_to_sync(self._recreate_db, True)
            self.initialize_db_tpl()

    @register_cli_command(name="init-tpl")
    def initialize_db_tpl(self):
        if self._settings.env not in ["dev", "pytest"]:
            TyperPrint.error(f"command forbidden in '{self._settings.env}' environment")
            raise typer.Exit()
        TyperPrint.info("Init Db Templates")
        AsyncUtils.async_to_sync(self._recreate_db_tpl)

    @register_cli_command(name="migrate")
    def migrate(self):
        self._do_migrations()

    @register_cli_command(name="makemigrations")
    def make_migrations(self, domain_name: str, migration_name: str):
        migration_dir = self._get_migrations_dir(domain_name)
        self._path_service.create_missing(migration_dir)
        os.system(f"yoyo new --sql {migration_dir}" f" -m '{migration_name}'")

    async def _recreate_db(self, db_test=False):
        db_name = (
            self._settings.db.name if not db_test else self._settings.db.db_test_name
        )
        settings = self._settings.db.model_copy(deep=True, update={"name": "template1"})
        provider = AsyncPgProvider(settings)
        async with provider.connection() as conn:
            await provider.disconnect_all(conn)
            # noinspection TryExceptPass,PyBroadException
            try:
                await self._drop_db(conn, db_name)
            except Exception:
                pass
            await self._create_db(conn, db_name)

    async def _recreate_db_tpl(self):
        for context in self._settings.bounded_contexts:
            await self._restore_db_tpl(DbTplUtils(context, self._settings))

    @staticmethod
    async def _create_db(conn, db_name):
        TyperPrint.info(f"  creating db '{db_name}' ...", with_newline=False)
        try:
            # noinspection SqlInjection
            sql = f"CREATE DATABASE {db_name}"
            await conn.execute(sql)
        except Exception as e:
            TyperPrint.error(f"  {e}")
            raise typer.Exit()
        TyperPrint.success(" OK")

    @staticmethod
    async def _drop_db(conn, db_name):
        TyperPrint.info(f"  droping db '{db_name}' ...", with_newline=False)
        try:
            # noinspection SqlInjection
            sql = f"DROP DATABASE {db_name}"
            await conn.execute(sql)
        except Exception as e:
            TyperPrint.error(f"  {e}")
            raise typer.Exit()
        TyperPrint.success(" OK")

    def _do_migrations(self):
        for context in self._settings.bounded_contexts:
            self._migrate_context(context)

    def _migrate_context(self, context):
        migrations_path = str(self._settings.get_migrations_path(context))
        TyperPrint.info(
            f"  applying migration(s) for {context} ...", with_newline=False
        )
        MigrationUtils.migrate(self._settings.db.dsn, migrations_path)
        TyperPrint.success(" OK")



import subprocess
from abc import abstractmethod

import typer
from asyncpg import InvalidCatalogNameError
from injector import inject

from griff.infra.cli.cli_controller import (
    CliController,
)
from griff.infra.cli.typer_print import TyperPrint
from griff.settings.griff_settings import GriffSettings
from griff.utils.db_tpl_utils import DbTplUtils


class AbstractDbCliController(CliController):
    @inject
    def __init__(self, settings: GriffSettings) -> None:
        self._settings = settings
        super().__init__()

    @abstractmethod
    def get_command_name(self) -> str:
        raise NotImplementedError

    async def _restore_db_tpl(self, db_tpl: DbTplUtils) -> None:
        TyperPrint.info(f"  restoring db {db_tpl.tpl_db_name} ...", with_newline=False)
        if db_tpl.dump_filename.exists() is False:
            TyperPrint.warning(" no db tpl found")
            return None

        try:
            await db_tpl.delete()
        except InvalidCatalogNameError:
            pass
        except Exception as e:
            TyperPrint.error(f"  {e}")
            raise typer.Exit()

        try:
            await db_tpl.create()
        except Exception as e:
            TyperPrint.error(f"  {e}")
            raise typer.Exit()

        if self._settings.env != "pytest":
            cmd = (
                f"PGPASSWORD={self._settings.db.password} "
                f"pg_restore -Fc -j 8 -U {self._settings.db.user} "
                f"-h {self._settings.db.host} "
                f"-p {self._settings.db.port} "
                f"-d {db_tpl.tpl_db_name} {db_tpl.dump_filename}"
            )
            p = subprocess.run(cmd, shell=True)  # noqa
            self._check_process_run(p, "restore")
        else:
            try:
                await db_tpl.restore_from_sql()
            except Exception as e:
                TyperPrint.error(f"  {e}")
                raise typer.Exit()

        TyperPrint.success(" OK")



from injector import inject

from griff.appli.app_event.app_event import AppEvent
from griff.appli.app_event.app_event_bus import AppEventBus
from griff.appli.command.command import Command
from griff.appli.command.command_handler import CommandResponse
from griff.appli.command.command_middleware import CommandMiddleware
from griff.appli.message.message_middleware import MessageContext


class CommandAppEventDispatchMiddleware(CommandMiddleware):
    @inject
    def __init__(self, app_event_bus: AppEventBus):
        super().__init__()
        self._app_event_bus = app_event_bus

    async def dispatch(
        self, message: Command, context: MessageContext | None = None
    ) -> CommandResponse:
        response = await self._next_dispatch(message, context)
        if response.is_success:
            for event in response.events:
                if isinstance(event, AppEvent):
                    await self._app_event_bus.dispatch(event, context)
        return response



from abc import ABC
from typing import TypeVar, Generic, Type, Any

from griff.appli.event.event import Event
from griff.appli.message.message_handler import (
    MessageHandler,
)
from griff.infra.registry.meta_registry import (
    MetaEventHandlerRegistry,
)

EM = TypeVar("EM", bound=Event)


class EventHandler(
    Generic[EM], MessageHandler[EM, None], ABC, metaclass=MetaEventHandlerRegistry
):
    ...


class FakeEventHandler(EventHandler, ABC):
    on_event_type: Type[Event]

    def __init__(self):
        super().__init__()
        self._log = {}

    async def handle(self, message: Event) -> None:
        self._log[message.event_name] = message.model_dump()  # type: ignore

    def list_events_handled(self) -> dict[str, Any]:
        return self._log

    @classmethod
    def listen_to(cls) -> Type[Event]:
        return cls.on_event_type



from fastapi import Request
from loguru import logger
from starlette import status
from starlette.responses import JSONResponse

from griff.infra.fastapi.exception_handlers.validation_exception_simplifier import (
    simplify_errors,
)
from griff.utils.errors import (
    get_error_by_status_code,
)
from griff.utils.exceptions import GriffException


async def griff_exception_handler(
    request: Request, exception: GriffException
) -> JSONResponse:  # pragma: no cover
    status_code = (
        exception.status_code
        if hasattr(exception, "status_code")
        else status.HTTP_500_INTERNAL_SERVER_ERROR
    )
    error_class = get_error_by_status_code(status_code)
    details = exception.details
    if status_code == 422:
        details = simplify_errors(details or [])  # type: ignore
    error = error_class(
        name=exception.short_classname().replace("Exception", "Error"),
        message=exception.message,
        details=details,
    )
    logger.info(f"GriffException: {error}")
    return JSONResponse(status_code=error.code, content=error.model_dump())



from abc import ABC

from griff.runtime.factories.abstract_runtime_factory import Injectables
from griff.test_utils.mixins.runtime_test_mixin import RuntimeTestMixin
from griff.test_utils.pytest_runtime_factory import PytestRuntimeFactory
from griff.test_utils.testcases.testcase import TestCase


class DomainTestCase(RuntimeTestMixin, TestCase, ABC):
    @classmethod
    def with_injectables(cls) -> Injectables | None:
        pass

    @classmethod
    def runtime_factory(cls) -> PytestRuntimeFactory:
        return super().runtime_factory().domain_test()



from argon2 import PasswordHasher
from argon2.exceptions import VerificationError, InvalidHashError
from injector import singleton

from griff.services.hasher.providers.hasher_provider import (
    HasherProvider,
    UnHashedStr,
    HashedStr,
)


@singleton
class Argon2Provider(HasherProvider):
    def __init__(self):
        self.ph = PasswordHasher()

    def hash(self, to_hash: UnHashedStr) -> HashedStr:
        return self.ph.hash(to_hash)

    def verify(self, candidate: UnHashedStr, reference: HashedStr) -> bool:
        try:
            return self.ph.verify(reference, candidate)
        except (VerificationError, InvalidHashError):
            return False



from typing import List

from griff.services.query_runner.query_runner_service import QueryRunnerService
from griff.test_utils.mixins.stub_mixin import StubMixin


class FakeQueryRunnerService(StubMixin, QueryRunnerService):
    async def run_query(self, query_name, **query_params):  # pragma: no cover
        return self._call_stub("run_query")

    async def run_many_query(
        self, query_name, queries_params: List[dict]
    ):  # pragma: no cover
        return self._call_stub("run_many_query")

    def set_sql_queries(self, relative_sql_path: str) -> None:  # pragma: no cover
        self._call_stub("set_sql_queries")



import copy


def simplify_errors(errors: list) -> dict:
    prepared_errors = _prepare_errors(errors)
    simplified_errors = {"errors": prepared_errors["errors"]}
    if prepared_errors["attr_errors"]:
        simplified_errors["attr_errors"] = _simplify_errors(
            prepared_errors["attr_errors"]
        )
    if prepared_errors["list_errors"]:
        simplified_errors["list_errors"] = _simplify_errors(
            prepared_errors["list_errors"]
        )
    return simplified_errors


def _simplify_errors(attr_errors):
    simplified_errors = {}
    for attr, error in attr_errors.items():
        simplified_errors[attr] = {}
        if not attr.isdigit() and not error["attr_errors"] and not error["list_errors"]:
            simplified_errors[attr] = error["errors"][0]
            continue

        simplified_errors[attr] = {"errors": error["errors"]}

        if error["attr_errors"]:
            simplified_errors[attr]["attr_errors"] = _simplify_errors(
                error["attr_errors"]
            )
        if error["list_errors"]:
            simplified_errors[attr]["list_errors"] = _simplify_errors(
                error["list_errors"]
            )
    return simplified_errors


def _prepare_errors(errors: list) -> dict:
    loc_error: dict = {"errors": [], "attr_errors": {}, "list_errors": {}}
    prepared_errors = copy.deepcopy(loc_error)
    for error in errors:
        dest = prepared_errors
        attr = []
        loc = [f for f in error["loc"] if f != "body"]
        if not loc:
            # global error
            dest["errors"].append(error["msg"])
            continue

        for loc_attr in loc:
            if not isinstance(loc_attr, int):
                attr.append(loc_attr)
                continue

            # attr is a list
            loc_attr = str(loc_attr)
            flat_attr = ".".join(attr)
            if flat_attr:
                if flat_attr not in dest["attr_errors"]:
                    dest["attr_errors"][flat_attr] = copy.deepcopy(loc_error)
                if loc_attr not in dest["attr_errors"][flat_attr]["list_errors"]:
                    dest["attr_errors"][flat_attr]["list_errors"][
                        loc_attr
                    ] = copy.deepcopy(loc_error)
                dest = dest["attr_errors"][flat_attr]["list_errors"][loc_attr]
            else:
                if loc_attr not in dest["list_errors"]:
                    dest["list_errors"][loc_attr] = copy.deepcopy(loc_error)
                dest = dest["list_errors"][loc_attr]
            attr = []

        if attr:
            flat_attr = ".".join(attr)
            if flat_attr not in dest["attr_errors"]:
                dest["attr_errors"][flat_attr] = copy.deepcopy(loc_error)
            dest["attr_errors"][flat_attr]["errors"].append(error["msg"])
        else:
            dest["errors"].append(error["msg"])
    return prepared_errors



from abc import ABC
from typing import TypeVar, Generic, List, Any, Literal

from pydantic import Field
from pydantic.main import IncEx

from griff.appli.command.command import Command
from griff.appli.event.event import Event
from griff.appli.message.message_handler import (
    MessageHandler,
    MessageErrorResponse,
    MessageSuccessResponse,
    EntityIdContent,
)
from griff.infra.registry.meta_registry import MetaCommandHandlerRegistry


class CommandSuccessResponse(MessageSuccessResponse, ABC):
    events: List[Event] = Field(default_factory=list)

    def model_dump(
        self,
        *,
        mode: Literal["json", "python"] | str = "python",
        include: IncEx | None = None,
        exclude: IncEx | None = None,
        context: Any | None = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        round_trip: bool = False,
        warnings: bool | Literal["none", "warn", "error"] = True,
        serialize_as_any: bool = False
    ) -> dict[str, Any]:
        result = super().model_dump(
            mode=mode,
            include=include,
            exclude=exclude,
            context=context,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
            serialize_as_any=serialize_as_any,
        )
        # pydantic dump model as only Event ignoring specific event fields
        result["events"] = [e.model_dump() for e in self.events]
        return result


class CommandErrorResponse(MessageErrorResponse, ABC):
    ...


class CommandEntityIdResponse(CommandSuccessResponse):  # pragma: no cover
    content: EntityIdContent


CommandResponse = CommandSuccessResponse | CommandErrorResponse

CM = TypeVar("CM", bound=Command)
CR = TypeVar("CR", bound=CommandResponse)


class CommandHandler(
    Generic[CM, CR], MessageHandler[CM, CR], ABC, metaclass=MetaCommandHandlerRegistry
):
    ...



from abc import ABC, abstractmethod

from griff.appli.message.message_middleware import MessageMiddleware, MessageContext
from griff.appli.query.query import Query
from griff.appli.query.query_handler import QueryResponse


class QueryMiddleware(MessageMiddleware[Query, QueryResponse], ABC):
    @abstractmethod
    async def dispatch(
        self, message: Query, context: MessageContext | None = None
    ) -> QueryResponse:
        pass



from injector import inject

from griff.appli.app_event.app_event import AppEvent
from griff.appli.app_event.app_event_dispatcher import AppEventDispatcher
from griff.appli.message.message_bus import MessageBus
from griff.appli.app_event.app_event_handler import AppEventHandler


class AppEventBus(MessageBus[AppEvent, None, AppEventHandler]):
    @inject
    def __init__(self, dispatcher: AppEventDispatcher) -> None:
        super().__init__(dispatcher)  # pragma: no cover



from abc import ABC
from typing import Annotated

from pydantic import BaseModel, ConfigDict, Field

EntityId = Annotated[str, Field(max_length=36, min_length=26, frozen=True)]


class DTO(BaseModel, ABC):
    ...


class ValueObject(BaseModel, ABC):
    model_config = ConfigDict(
        use_enum_values=True, validate_default=True, extra="forbid", frozen=True
    )


class Entity(BaseModel, ABC):
    model_config = ConfigDict(use_enum_values=True, validate_default=True, frozen=True)
    entity_id: EntityId

    @classmethod
    def classname(cls) -> str:  # pragma: no cover
        return str(cls)

    @classmethod
    def short_classname(cls) -> str:
        return cls.__name__

    @classmethod
    def _entity_label(cls) -> str:
        return cls.short_classname()


# noinspection Pydantic
class Aggregate(Entity, ABC):
    pass



import asyncio
from typing import Any, Callable


class AsyncUtils:
    @staticmethod
    def async_to_sync(async_func: Callable, args: object = None) -> Any:
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:  # pragma: no cover
            loop = asyncio.new_event_loop()

        if isinstance(args, dict):  # pragma: no cover
            coroutine = async_func(**args) if args else async_func()
        elif isinstance(args, list):  # pragma: no cover
            coroutine = async_func(*args) if args else async_func()
        else:
            coroutine = async_func(args) if args else async_func()

        return loop.run_until_complete(coroutine)



from abc import ABC, abstractmethod
from typing import Generic, Type, TypeVar, Self, Any, Literal

from pydantic import BaseModel, model_validator

from griff.appli.message.message import Message
from griff.domain.common_types import EntityId
from griff.utils.errors import BaseError


class AbstractMessageResponse(BaseModel, ABC):
    code: int = 200
    content: Any | None = None
    is_success: bool

    @property
    def is_failure(self) -> bool:  # pragma: no cover
        return not self.is_success

    def dump_content(self) -> BaseModel | None:  # pragma: no cover
        if self.content:
            return self.content.model_dump()
        return self.content


class MessageSuccessResponse(AbstractMessageResponse):
    content: Any | None = None
    is_success: Literal[True] = True


class EntityIdContent(BaseModel):  # pragma: no cover
    entity_id: EntityId


class MessageErrorResponse(AbstractMessageResponse):
    code: int = 500
    content: BaseError
    is_success: Literal[False] = False

    @model_validator(mode="after")
    def set_code_from_error_code(self) -> Self:
        self.code = self.content.code
        return self


MessageResponse = MessageSuccessResponse | MessageErrorResponse | None
M = TypeVar("M", bound=Message)
MR = TypeVar("MR", bound=MessageResponse)


class MessageHandler(Generic[M, MR], ABC):
    @abstractmethod
    async def handle(self, message: M) -> MR:  # pragma: no cover
        pass

    @classmethod
    def handlers(cls):
        return cls.__subclasses__()

    @classmethod
    @abstractmethod
    def listen_to(cls) -> Type[M]:  # pragma: no cover
        pass



from abc import ABC, abstractmethod
from typing import List, Any, Dict

QueryRowResult = Dict[str, Any]
QueryRowResults = List[QueryRowResult]
QueryResult = None | QueryRowResult | QueryRowResults


class Persistence(ABC):  # pragma: no cover
    async def insert(self, data: dict) -> None:
        prepared_data = self._prepare_to_save(data)
        return await self._insert(prepared_data)

    async def update(self, data: dict) -> None:
        prepared_data = self._prepare_to_save(data)
        return await self._update(prepared_data)

    async def delete(self, persistence_id: str) -> None:
        return await self._delete(persistence_id)

    async def get_by_id(self, persistence_id: str) -> QueryRowResult | None:
        result = await self._get_by_id(persistence_id)
        return self._prepare_row_result(result) if result else None

    async def list_all(self) -> QueryRowResults:
        results = await self._list_all()
        return self._prepare_row_results(results)

    async def run_query(self, query_name: str, **query_params) -> QueryResult:
        results = await self._run_query(query_name, **query_params)
        if isinstance(results, list):
            return self._prepare_row_results(results)
        return self._prepare_row_result(results) if results else None  # type: ignore

    @abstractmethod
    async def _insert(self, data: dict) -> None:
        ...

    @abstractmethod
    async def _update(self, data: dict) -> None:
        ...

    @abstractmethod
    async def _delete(self, persistence_id: str) -> None:
        ...

    @abstractmethod
    async def _get_by_id(self, persistence_id: str) -> QueryRowResult | None:
        ...

    @abstractmethod
    async def _list_all(self) -> QueryRowResults:
        ...

    @abstractmethod
    async def _run_query(self, query_name: str, **query_params) -> QueryResult:
        ...

    def reset(self, initial_data: List[Dict] | None = None):
        # only for testing purposes
        return None

    # noinspection PyMethodMayBeStatic
    def _prepare_to_save(self, data: dict) -> dict:
        return data

    # noinspection PyMethodMayBeStatic
    def _prepare_row_result(self, result: QueryRowResult) -> QueryRowResult:
        return result

    def _prepare_row_results(self, results: QueryRowResults) -> QueryRowResults:
        return [self._prepare_row_result(row) for row in results]



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from injector import inject

from griff.domain.common_types import Aggregate
from griff.infra.persistence.serialized_persistence import SerializedPersistence
from griff.infra.repository.repository import Repository
from griff.services.date.date_service import DateService

A = TypeVar("A", bound=Aggregate)


class SerializedRepository(Generic[A], Repository[A], ABC):
    @inject
    @abstractmethod
    def __init__(
        self,
        persistence: SerializedPersistence,
        date_service: DateService,
    ):
        super().__init__(persistence, date_service)



from decimal import Decimal
from typing import Any

from pydantic import model_validator

from griff.domain.common_types import ValueObject


class Money(ValueObject):
    amount: Decimal
    currency: str = "EUR"

    @model_validator(mode="before")
    @classmethod
    def check_card_number_omitted(cls, data: Any) -> Any:
        if isinstance(data, dict) is False:
            data = {"amount": str(data)}
        return data

    def __str__(self):  # pragma: no cover
        return f"{self.amount}{self.currency}"

    def __ge__(self, other):
        self._check_currency_match(other)
        return self.amount >= other.amount

    def __le__(self, other):
        self._check_currency_match(other)
        return self.amount <= other.amount

    def __gt__(self, other):
        self._check_currency_match(other)
        return self.amount > other.amount

    def __lt__(self, other):
        self._check_currency_match(other)
        return self.amount < other.amount

    def _check_currency_match(self, other):
        if self.currency != other.currency:
            raise ValueError("Currency mismatch")



from abc import ABC
from typing import Any

from starlette.responses import JSONResponse

from griff.appli.message.message_handler import (
    MessageSuccessResponse,
    MessageErrorResponse,
)
from griff.appli.message.message_middleware import MessageContext
from griff.infra.registry.meta_endpoint_controller_registry import (
    MetaEndpointControllerRegistry,
)
from griff.utils.inspect_utils import find_bound_method_to_object


class ApiController(ABC, metaclass=MetaEndpointControllerRegistry):
    def __init__(self) -> None:
        """
        Auto register endpoint in fastApi router
        """
        endpoint_list = MetaEndpointControllerRegistry.get_endpoint_registry()[
            type(self)
        ]
        self._endpoints = list()
        for endpoint in endpoint_list:
            endpoint_route = MetaEndpointControllerRegistry.get_full_route_for_endpoint(
                controller=self, endpoint=endpoint
            )
            if endpoint_route is None:  # pragma: no cover
                continue
            self._endpoints.append(
                {
                    "route": endpoint_route,
                    "method": endpoint.http_method,
                    "func": find_bound_method_to_object(self, endpoint.endpoint),
                    "return_code": endpoint.http_success_code,
                }
            )

    def get_endpoints(self) -> list[dict[str, Any]]:
        return self._endpoints

    @staticmethod
    def prepare_response(
        response: MessageSuccessResponse | MessageErrorResponse,
    ) -> Any | JSONResponse:  # pragma: no cover
        if response.is_success:
            return response.content

        return JSONResponse(status_code=response.code, content=response.dump_content())

    @staticmethod
    def _get_token_context(token: str) -> MessageContext:  # pragma: no cover
        return MessageContext(context={"access_token": token})



from griff.services.date.date_models import DateTime
from griff.services.date.date_service import DateService


class FakeDateService(DateService):
    def __init__(self):
        super().__init__()
        self._today = "2023-04-08 11:11:11.111111"

    def set_today(self, today: str):  # pragma: no cover
        self._today = today

    def now(self) -> DateTime:
        return DateTime(self._today)



from abc import ABC
from typing import Dict, Generic, Self, TypeVar

from griff.appli.message.message_handler import (
    Message,
    MessageHandler,
    MessageResponse,
)
from griff.appli.message.message_middleware import MessageMiddleware, MessageContext

M = TypeVar("M", bound=Message)
MR = TypeVar("MR", bound=MessageResponse | None)
MH = TypeVar("MH", bound=MessageHandler)
HandlerList = Dict[str, MH]


class MessageDispatcher(Generic[M, MR], MessageMiddleware[M, MR], ABC):
    def __init__(self) -> None:
        super().__init__()
        self._handlers: HandlerList = {}

    async def dispatch(self, message: M, context: MessageContext | None = None) -> MR:
        if message.classname() not in self._handlers:
            raise RuntimeError(
                f"No handler registered for '{message.short_classname()}'"
            )
        response = await self._handlers[message.classname()].handle(message)
        return response

    def set_next(self, middleware: MessageMiddleware) -> MessageMiddleware:
        raise RuntimeError(f"{self.__class__.__name__} must be the last middleware")

    def register(self, handlers: list[MH]) -> Self:
        for handler in handlers:
            self._register(handler)
        return self

    def _register(self, handler: MessageHandler) -> Self:
        if handler.listen_to().classname() in self._handlers:
            raise RuntimeError(
                f"Handler '{handler.__class__.__name__}' already registered"
            )
        # noinspection PyTypeChecker
        self._handlers[handler.listen_to().classname()] = handler
        return self



import enum

from pydantic import BaseModel


class JwtAlgorithm(str, enum.Enum):
    hs256 = "HS256"


class JwtSettings(BaseModel):
    access_secret_key: str
    refresh_secret_key: str
    access_token_lifetime_seconds: int = 5 * 60  # 5 min
    refresh_token_lifetime_seconds: int = 60 * 60 * 24 * 7  # 7 days
    algorithm: JwtAlgorithm = JwtAlgorithm.hs256



import typing
from pathlib import Path

from fastapi import FastAPI
from fastapi.exceptions import RequestValidationError
from fastapi_babel import BabelConfigs, BabelMiddleware
from injector import Binder, singleton, Injector
from loguru import logger
from pydantic_i18n import BabelLoader, PydanticI18n
from starlette.middleware.cors import CORSMiddleware

from griff.infra.fastapi.exception_handlers.griff_exception_handler import (
    griff_exception_handler,
)
from griff.infra.fastapi.exception_handlers.pydantic_exception_handler import (
    PydanticValidationExceptionHandler,
)
from griff.runtime.components.abstract_runtime_component import (
    RuntimeComponent,
    InjectBindable,
    Runnable,
)
from griff.settings.griff_settings import GriffSettings
from griff.utils.exceptions import GriffException


class FastApiRuntimeComponent(Runnable, InjectBindable, RuntimeComponent):
    def __init__(self, settings: GriffSettings):
        self._settings = settings

    def configure(self, binder: Binder) -> None:
        binder.bind(FastAPI, to=FastAPI, scope=singleton)

    # noinspection PyTypeChecker
    @typing.no_type_check
    def initialize(self, injector: Injector):
        app = injector.get(FastAPI)
        app.add_middleware(CORSMiddleware, **self._settings.cors.model_dump())
        app.add_exception_handler(GriffException, griff_exception_handler)
        self.init_fastapi_locale(app)
        self.init_pydantic_locale_and_exception_handler(app)

    def clean(self, injector: Injector):
        pass

    def start(self, injector: Injector):
        pass

    def stop(self, injector: Injector):
        pass

    def init_fastapi_locale(self, app: FastAPI):
        if self._settings.locale is None:
            logger.warning("No locale settings found. Skipping locale initialization.")
            return None

        # # check translation messages are compiled
        translation_filename = Path(self._settings.project_dir).joinpath(
            f"{self._settings.fastapi_local_path}/fr/LC_MESSAGES/messages.mo"
        )
        if translation_filename.exists() is False:
            raise RuntimeError(
                "Translation files are missing, run "
                "'python manage.py locale compilemessages'"
            )

        configs = BabelConfigs(
            ROOT_DIR=f"{self._settings.project_dir}/app",  # bug ? config do a .parent
            BABEL_DEFAULT_LOCALE=self._settings.locale.default,
            BABEL_TRANSLATION_DIRECTORY=self._settings.locale.relative_path,
        )
        app.add_middleware(BabelMiddleware, babel_configs=configs)

    def init_pydantic_locale_and_exception_handler(self, app):
        if self._settings.locale is None:
            return None
        loader = BabelLoader(f"{self._settings.pydantic_locale_path}")
        pydantic_i18n = PydanticI18n(
            loader, default_locale=self._settings.locale.default
        )
        pydantic_exc_handler = PydanticValidationExceptionHandler(pydantic_i18n)
        app.add_exception_handler(RequestValidationError, pydantic_exc_handler.dispatch)



from typing import Dict


class CliDecoratorProcessor:
    def __init__(self) -> None:
        self._name: str | None = None

    def _sanitize(self, to_be_sanitized: str) -> str:
        return to_be_sanitized.strip("'").strip('"').strip(" ")

    def _process_name_param(self, a_str_to_process: str) -> None:
        splited = a_str_to_process.split("=")
        self._name = self._sanitize(splited[1])

    def load_decorators(self, a_decoration_string_list: list[str]) -> None:
        for decorator in a_decoration_string_list:
            if decorator[:21] == "@register_cli_command":
                keyval_list = decorator[
                    decorator.index("(") + 1 : decorator.index(")")  # noqa: E203
                ].split(",")
                for decoration_param in keyval_list:
                    if "=" not in decoration_param:
                        raise ValueError("decoration badly formatted, '=' not found")
                    self._process_name_param(a_str_to_process=decoration_param)

    def get_decorator_params(self) -> Dict[str, str | None]:
        return {"name": self._name}

    @property
    def name(self) -> str | None:
        return self._name



from typing import Dict, Type, TypeVar

from griff.services.abstract_service import AbstractService

S = TypeVar("S", bound=AbstractService)
ServiceContainer = Dict[str, S]


class ServiceLocator:
    _container: ServiceContainer = {}

    def __new__(cls, *args, **kwargs):
        raise RuntimeError(f"{cls.__name__} can not be instantiated")

    @classmethod
    def register(cls, service_class: Type[S], service: S) -> None:
        container_key = cls._get_container_key(service_class)
        if container_key not in cls._container:
            cls._container[container_key] = service
            return None
        raise RuntimeError(f"{service.__class__.__name__} is already registered")

    @classmethod
    def get(cls, service_class: Type[S]) -> S:
        container_key = cls._get_container_key(service_class)
        if container_key in cls._container:
            return cls._container[container_key]
        raise RuntimeError(f"{service_class.__name__} is not registered")

    @classmethod
    def reset(cls):
        cls._container = {}

    @classmethod
    def _get_container_key(cls, service: Type[S]) -> str:
        return str(service)



import threading
from contextlib import asynccontextmanager
from typing import List

import asyncpg
from asyncpg import Pool, Connection
from asyncpg.transaction import Transaction
from injector import inject, singleton
from loguru import logger

from griff.services.db.db_settings import DbSettings
from griff.services.db.providers.db_provider import DbProvider, DbRow


@singleton
class AsyncPgProvider(DbProvider[Connection]):
    @inject
    def __init__(self, settings: DbSettings):
        self._settings = settings
        self._transactions: dict[int, Transaction] = {}
        self._pool: Pool | None = None
        self._debug = False

    async def start(self) -> None:
        if self._pool is None:
            self.debug("--->START POOL")
            self._pool = await asyncpg.create_pool(
                dsn=self._settings.dsn, **self._settings.pool.model_dump()
            )

    async def stop(self) -> None:
        if self._pool is not None:
            self.debug("--->STOP POOL")
            self._pool.terminate()
            self._pool = None

    async def get_connection(self) -> Connection:
        if self._pool:
            self.debug("--->GET CONNECTION FROM POOL")
            return await self._pool.acquire()
        self.debug("--->GET CONNECTION")
        return await asyncpg.connect(self._settings.dsn)

    async def close_connection(self, connection: Connection) -> None:
        if self._pool:
            self.debug(f"--->RELEASE POOL CONNECTION {id(connection)}")
            await self._pool.release(connection)
            return None
        self.debug(f"--->CLOSE CONNECTION {id(connection)}")
        await connection.close()

    @asynccontextmanager
    async def connection(self) -> Connection:
        connection = await self.get_connection()
        try:
            yield connection
        finally:
            await self.close_connection(connection)

    async def execute(self, connection: Connection, sql: str | List[str]) -> None:
        if isinstance(sql, list):
            sql = ";\n".join(sql)
        self.debug(f"--->execute {sql}")
        await connection.execute(sql)

    async def fetch_one(self, connection: Connection, sql: str) -> DbRow:
        self.debug(f"--->fetch_one {sql}")
        row = await connection.fetchrow(sql)
        return dict(row)

    async def fetch_all(self, connection: Connection, sql: str) -> List[DbRow]:
        self.debug(f"--->fetch_all {sql}")
        rows = await connection.fetch(sql)
        return [dict(row) for row in rows]

    async def start_transaction(self, connection: Connection) -> None:
        conn_id = self._get_connection_id(connection)
        if conn_id in self._transactions:
            raise RuntimeError(
                f"a transaction is already started for connection {conn_id}"
            )

        self._transactions[conn_id] = connection.transaction()
        await self._transactions[conn_id].start()

    async def commit_transaction(self, connection: Connection) -> None:
        conn_id = self._get_connection_id(connection)
        transaction = self._check_transaction_exists(conn_id)
        await transaction.commit()
        self._transactions.pop(conn_id)

    async def rollback_transaction(self, connection: Connection) -> None:
        conn_id = self._get_connection_id(connection)
        transaction = self._check_transaction_exists(conn_id)
        await transaction.rollback()
        self._transactions.pop(conn_id)

    @asynccontextmanager
    async def transaction(self, connection: Connection) -> Transaction:
        async with connection.transaction():
            yield

    async def disconnect_all(self, connection: Connection) -> None:
        # noinspection SqlInjection
        sql = (
            "SELECT pg_terminate_backend(pg_stat_activity.pid) "
            "FROM pg_stat_activity "
            f"WHERE pg_stat_activity.datname = '{self._settings.name}' "
            "AND pid <> pg_backend_pid();"
        )
        await connection.execute(sql)

    def _check_transaction_exists(self, conn_id: int) -> Transaction:
        if conn_id in self._transactions:
            return self._transactions[conn_id]
        raise RuntimeError(f"no transaction started for connection {conn_id}")

    @staticmethod
    def _get_connection_id(connection: Connection) -> int:
        return id(connection)

    def debug(self, msg: str) -> None:  # pragma: no cover
        if self._debug is True:
            logger.debug(f" PG : {threading.current_thread().ident} {msg}")



from loguru import logger

from griff.appli.event.event import Event
from griff.appli.event.event_middleware import EventMiddleware
from griff.appli.message.message_middleware import MessageContext


class AppEventLoggerMiddleware(EventMiddleware):
    async def dispatch(
        self, message: Event, context: MessageContext | None = None
    ) -> None:  # pragma: no cover
        logger.info(f"dispatch app event: {message.short_classname()}")
        logger.debug(message.model_dump())
        if self._next:
            await self._next_dispatch(message, context)



import typing
from abc import ABC
from typing import Any, Type

from pydantic import BaseModel, model_validator


class BaseError(BaseModel, ABC):  # pragma: no cover
    code: int
    name: str | None = None
    message: str
    details: dict | None = None

    @model_validator(mode="before")
    @classmethod
    def set_name_from_classname(cls, data: Any) -> str:
        if isinstance(data, dict) and "name" not in data:
            data["name"] = cls.__name__
        return data

    @classmethod
    def short_classname(cls) -> str:
        return cls.__name__


class UnauthorizedError(BaseError, ABC):  # pragma: no cover
    code: int = 401
    message: str = "unauthorized access"


class AccessForbiddenError(BaseError, ABC):  # pragma: no cover
    code: int = 403
    message: str = "access forbidden"


class NotFoundError(BaseError, ABC):  # pragma: no cover
    code: int = 404
    message: str = "not found"


class ConflictError(BaseError, ABC):  # pragma: no cover
    code: int = 409
    message: str = "conflict error"


class ValidationError(BaseError, ABC):  # pragma: no cover
    code: int = 422
    message: str = "validation error"


class InternalError(BaseError, ABC):  # pragma: no cover
    code: int = 500
    message: str = "internal error"


class DefaultExceptionHandlerError(BaseError):
    pass


Error = (
    UnauthorizedError
    | AccessForbiddenError
    | NotFoundError
    | ConflictError
    | ValidationError
    | InternalError
)


@typing.no_type_check
def get_error_by_status_code(status_code: int) -> Type[Error]:
    return {
        401: UnauthorizedError,
        403: AccessForbiddenError,
        404: NotFoundError,
        409: ConflictError,
        422: ValidationError,
        500: InternalError,
    }[status_code]



from typing import Self, Generic

from griff.appli.event.event import Event
from griff.appli.message.message_dispatcher import MessageDispatcher, M, MR
from griff.appli.message.message_handler import MessageHandler
from griff.appli.message.message_middleware import MessageContext


class EventDispatcher(Generic[M, MR], MessageDispatcher[Event, None]):
    async def dispatch(
        self, message: Event, context: MessageContext | None = None
    ) -> None:
        if message.classname() not in self._handlers:
            return None

        for handler in self._handlers[message.classname()]:
            await handler.handle(message)

    def _register(self, handler: MessageHandler) -> Self:
        if handler.listen_to().classname() not in self._handlers:
            self._handlers[handler.listen_to().classname()] = []

        already_registered = any(
            handler.__class__ == h.__class__
            for h in self._handlers[handler.listen_to().classname()]
        )

        if already_registered:
            raise RuntimeError(
                f"Handler '{handler.__class__.__name__}' already registered"
            )

        self._handlers[handler.listen_to().classname()].append(handler)
        return self



from typing import Type, Any

from injector import Injector, Binder, singleton

from griff.appli.event.event_bus import EventBus
from griff.appli.event.event_handler import EventHandler
from griff.appli.event.event_middleware import EventMiddleware
from griff.appli.message.message_bus import ListMiddlewares
from griff.infra.registry.meta_registry import MetaEventHandlerRegistry
from griff.runtime.components.abstract_runtime_component import (
    RuntimeComponent,
    Runnable,
    InjectBindable,
    import_handlers,
)
from griff.settings.griff_settings import GriffSettings


class EventBusRuntimeComponent(Runnable, InjectBindable, RuntimeComponent):
    def __init__(
        self, settings: GriffSettings, middlewares: list[Type[EventMiddleware]]
    ):
        self._settings = settings
        self._middlewares = middlewares

    def configure(self, binder: Binder) -> None:
        binder.bind(EventBus, to=EventBus, scope=singleton)

    def initialize(self, injector: Injector) -> None:
        event_bus = injector.get(EventBus)
        middlewares: ListMiddlewares = [injector.get(m) for m in self._middlewares]
        for context in self._settings.bounded_contexts:
            import_handlers(
                self._settings.project_dir,
                self._settings.get_event_handlers_path(context),
            )
        handlers: list[EventHandler[Any]] = [
            injector.get(h) for h in MetaEventHandlerRegistry.list_types()
        ]
        event_bus.initialize(handlers, middlewares)

    def clean(self, injector: Injector) -> None:
        pass

    def start(self, injector: Injector) -> None:
        pass

    def stop(self, injector: Injector) -> None:
        pass



import datetime
import inspect
import json
import os
import subprocess
import tempfile
from abc import ABC
from decimal import Decimal
from pathlib import Path
from shutil import rmtree, which
from typing import Any

from pydantic import BaseModel, ValidationError

from griff.services.path.path_service import PathService


class AbstractTestCase(ABC):  # pragma: no cover
    _base_dir = None
    _base_test_dir = "/_tests"
    _fixture_dir = None
    _common_fixture_dir = None

    @classmethod
    def setup_class(cls) -> None:
        if hasattr(super(), "setup_class"):
            raise RuntimeError("Must be the last class in the inheritance chain")
        cls.clean_sandbox()
        cls.read_file = PathService().read_file  # type: ignore

    def setup_method(self) -> None:
        if hasattr(super(), "setup_method"):
            raise RuntimeError("Must be the last class in the inheritance chain")

    @classmethod
    def teardown_class(cls) -> None:
        pass

    def teardown_method(self) -> None:
        pass

    @classmethod
    def get_dataset(cls, dataset) -> Any:
        """
        Get fixture dataset if is json, load it else just read it
        Args:
            dataset(str): relative path to dataset directory
        Returns:
            mixed
        """
        return cls._get_dataset(dataset, cls.get_datasets_dir(dataset))

    @classmethod
    def get_datasets_dir(cls, relative_path=None) -> str:
        dir = f"{cls._get_test_fixtures_dir()}/datasets"
        if relative_path:
            return f"{dir}/{relative_path}"
        return dir

    def assert_equals_resultset(self, actual, **kwargs):
        # build resultset filename from caller filename
        calling_filename = kwargs.get("calling_filename", None)
        filename = self._get_calling_filename()
        calling_frame = self._get_calframe(calling_filename)
        method_name = self._get_calling_method_name(calling_frame)
        sub_dir = self._get_test_subdir(filename)
        filename_noext = Path(filename).name.replace(".py", "")
        resultset_filename = Path(
            f"{self._get_resultsets_dir()}/{sub_dir}/{filename_noext}"
            f"/{method_name}.json"
        )
        self._create_missing_directories(resultset_filename)

        actual_json = self._json_dump(actual)

        # get expected result sets from file
        if resultset_filename.exists() is False:
            # create missing file
            data = (
                actual_json
                if os.environ.get("TEST_RESULTSET_AUTOSAVE", False)
                else "{}"
            )
            with resultset_filename.open("w") as f:
                f.write(data)

        with resultset_filename.open() as json_file:
            expected = json_file.read()

        # assert

        try:
            assert expected == actual_json
        except AssertionError as e:
            tmp_dir = self._get_tmp_dir()

            # build fix result set
            tmp_actual_filename = Path(
                f"{tmp_dir}/{filename_noext}-{method_name}_ACTUAL.json"
            )
            self._create_missing_directories(tmp_actual_filename)

            with tmp_actual_filename.open("w") as f:
                f.write(actual_json)

            common_cmd = f"{resultset_filename} {tmp_actual_filename}"
            if which("bcompare") is not None:
                common_cmd = f"bcompare {common_cmd}"
            elif which("pycharm") is not None:
                common_cmd = f"pycharm diff {common_cmd}"
            elif which("charm") is not None:
                common_cmd = f"charm diff {common_cmd}"
            elif which("pycharm-community") is not None:
                common_cmd = f"pycharm-community diff {common_cmd}"
            elif which("meld") is not None:
                common_cmd = f"meld {common_cmd}"
            # noinspection SubprocessShellMode
            subprocess.Popen(common_cmd, shell=True)

            print(f"\n\033[1;31m=== TEST {method_name} has failed !!")
            print(f"file: {filename}")
            os.system(f"diff {resultset_filename} {tmp_actual_filename}")
            print(f"=== TEST {method_name}\n\033[0m")
            raise e

    @classmethod
    def clean_sandbox(cls):
        cls._empty_dir(cls._get_base_sandbox_dir())

    @classmethod
    def get_sandbox_dir(cls, filename=None) -> str:
        sandbox_path = cls._get_base_sandbox_dir().joinpath(cls.__name__)
        if filename is None:
            sandbox_path.mkdir(parents=True, exist_ok=True)
            return str(sandbox_path)
        filename = Path(f"{sandbox_path}/{filename}")
        cls._create_missing_directories(filename)
        return str(filename)

    @staticmethod
    def assert_path_exists(path):
        assert os.path.exists(path), path

    @staticmethod
    def assert_path_not_exists(path):
        assert not os.path.exists(path), path

    @classmethod
    def _empty_dir(cls, directory: Path) -> None:
        full_path = Path(directory)
        if full_path.exists():
            rmtree(full_path)
            full_path.mkdir()

    @classmethod
    def _get_dataset(cls, dataset: str, dataset_filename: str) -> Any:
        filename, file_extension = os.path.splitext(dataset)
        with open(dataset_filename) as file:
            if file_extension == ".json":
                data = json.load(file)
                if "__generated_by__" in dataset:
                    data.pop("__generated_by__")
                return data
            return file.read()

    @classmethod
    def _get_tmp_dir(cls) -> str:
        """Get tmp test directory"""
        return f"{tempfile.tempdir}/.donotcommit_tmp"

    @classmethod
    def _get_test_fixtures_dir(cls) -> str:
        """
        Get fixtures directory

        Returns:
            str: fixtures directory
        """
        if not cls._fixture_dir:
            filename = cls._get_calling_filename()
            pos = filename.find(cls._base_test_dir)
            if not pos > 0:
                raise ValueError(f"Unable to determine root tests path from {filename}")
            cls._fixture_dir = f"{filename[:pos]}{cls._base_test_dir}/_fixtures"

        return cls._fixture_dir

    @classmethod
    def _get_resultsets_dir(cls) -> str:
        """
        Get result sets dir

        Returns:
            str: result sets directory
        """
        return f"{cls._get_test_fixtures_dir()}/resultsets"

    @classmethod
    def _get_base_sandbox_dir(cls):
        return Path(f"{cls._get_test_fixtures_dir()}/sandbox")

    @classmethod
    def _get_calling_filename(cls) -> str:
        return inspect.getfile(cls)

    @classmethod
    def _get_calframe(cls, calling_filename=None) -> inspect.FrameInfo:
        # find calling frame
        calframe = inspect.getouterframes(inspect.currentframe(), 2)
        test_filename = (
            cls._get_calling_filename() if not calling_filename else calling_filename
        )
        for frame in calframe:
            if frame.filename.find(test_filename) >= 0:
                return frame
        raise ValueError("Unable to determine calling frame")

    @staticmethod
    def _get_calling_method_name(frame) -> Any:
        return frame[3]

    @staticmethod
    def _get_calling_lineno(frame) -> str:
        return frame.lineno

    @staticmethod
    def _create_missing_directories(filename: Path) -> None:
        if filename.exists():
            return None

        filename = filename.parent
        filename.mkdir(parents=True, exist_ok=True)

    @classmethod
    def _get_test_subdir(cls, filename):
        dir_names = str(Path(filename).parent).split("/")

        # search tests dir
        for _ in range(0, len(dir_names)):
            dir_name = dir_names.pop(0)
            if dir_name == "_tests":
                break

        return "/".join(dir_names)

    @classmethod
    def _json_dump(cls, data):
        json_dump = json.dumps(
            cls._to_json_dumpable(data),
            indent=4,
            separators=(",", ": "),
            default=str,
            # disable non-ASCII characters escape with \uXXXX sequences
            ensure_ascii=False,
        )
        # for respecting file empty last line convention
        return f"{json_dump}\n"

    @classmethod
    def _to_json_dumpable(cls, data):
        """
        Convert data to be json dumpable
        """
        if isinstance(data, BaseModel):
            data = data.model_dump()

        if isinstance(data, tuple):
            data = list(data)

        if isinstance(data, dict):
            jsonized = {}
            for k, v in data.items():
                new_k = cls._to_json_dumpable(k)
                jsonized[new_k] = cls._to_json_dumpable(v)
            return jsonized

        if isinstance(data, list):
            for i, v in enumerate(data):
                data[i] = cls._to_json_dumpable(v)
            return data

        if isinstance(data, set):
            data = cls._to_json_dumpable(list(data))

        if isinstance(data, Decimal):
            return float(data)

        # if isinstance(data, Arrow):
        #     return data.format()

        if isinstance(data, datetime.datetime):
            return data.strftime("%Y-%m-%d %H:%M:%S")

        if isinstance(data, datetime.date):
            return data.strftime("%Y-%m-%d")

        if isinstance(data, ValidationError):
            return cls._to_json_dumpable(data.errors(include_url=False))

        return data
